/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of doubly linked list using external storage model.
 * see http://en.wikipedia.org/wiki/Linked_list for detail.
 */
#ifndef _DLLIST_H
#define _DLLIST_H

#include <stddef.h>
#include <stdarg.h>
#include "ulbool.h"
#include "uldef.h"
_BEGIN_DECLS

typedef struct dll_node_s {
	const void *data;		/* const pointer to external storage. */
	struct dll_node_s *prev;	/* pointer to previous node. */
	struct dll_node_s *next;	/* pointer to next node. */
} dll_node_t;				/* ADT of DLL node. */

typedef struct {
	dll_node_t *head;		/* pointer to head node. */
	dll_node_t *tail;		/* pointer to tail node. */
	size_t linklen;			/* number of nodes DLL contains. */
} dll_t;				/* ADT of DLL. */


/* Init DLL structure to zero. */
void dll_init(dll_t *dll);

/* Duplicate another DLL, pointing to the same external storage. */
bool dll_dup(dll_t *dst, const dll_t *src);

/* Peek specific node's element content,
 * return the pointer on valid offset, otherwise NULL. */
const void *dll_node_peek(dll_t *dll, size_t offset);

/* Delete DLL's current head node. */
void dll_delhead(dll_t *dll);

/* Delete DLL's current tail node. */
void dll_deltail(dll_t *dll);

/* Scan link and update matching node(s),(test() returns true)
 * return number of updated node(s). */
size_t dll_node_cupdt(dll_t *dll, bool (*test)(const void *), const void *data);
	
/* Add a new head node. */
bool dll_addhead(dll_t *dll, const void *data);

/* Add a new tail node (append). */
bool dll_addtail(dll_t *dll, const void *data);

/* Conditionally (callback returns true) delete DLL node(s),
 * return number of deleted node(s). */
size_t dll_node_cdel(dll_t *dll, bool (*test)(const void *));

/* Insert a node right before matching node (callback returns true),
 * return true on success, otherwise false. */
bool dll_node_bins(dll_t *dll, bool (*test)(const void *), const void *data);

/* Insert a node right after matching node (callback returns true),
 * return true on success, otherwise false. */
bool dll_node_ains(dll_t *dll, bool (*test)(const void *), const void *data);

/* Return number of node(s) DLL contains. */
size_t dll_getlen(dll_t *dll);

/* Sort DLL content by exchanging const pointers to external storage. */
void dll_sort(dll_t *dll, int (*cmp)(const void *, const void *));

/* Merge asc/desc sorted links (all in same order) into a single sorted dst, 
 * various dll_t * parameters and a terminating NULL is expected. 
 * original ones will be destroyed automatically. */
void dll_merge(dll_t *dst, int (*cmp)(const void *, const void *), ...);

/* Traverse DLL via visit() from head to tail if flag is true,
 * else from tail to head. */
void dll_trav(dll_t *dll, bool flag, void (*visit)(const void *));

/* Free DLL allocated memory. */
void dll_destroy(dll_t *dll);

_END_DECLS
#endif

