/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of queue (FIFO), in 3 forms.
 * DYNAMIC: slowest but flexible, handy when elem structure is really big.
 * STATIC: fastest but must know maximum size/len of queue.
 * BALANCED: general purpose circular queue, balanced performance/memory usage.
 * see http://en.wikipedia.org/wiki/Queue_(data_structure) for detail.
 */
#if defined(ADT_DEBUG) || defined(ADT_QUEUE_DEBUG)
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include "queue.h"

/* Init queue ADT, call this before any following operations. */
void queue_init(queue_t *p)
{
#ifdef ADT_QUEUE_BALANCED
	p->rear = p->front = p->memstart = NULL;
	p->memunits = 0;
#endif
#ifdef ADT_QUEUE_DYNAMIC
	sll_init(p);
#endif
#ifdef ADT_QUEUE_STATIC
	p->front = p->rear = p->e;
#endif
}

/* Attach a node to queue and make it new rear.
 * return false on malloc() failure, else true. */
bool queue_attach(queue_t *p, const void *data)
{
#ifdef ADT_QUEUE_BALANCED
	if(p->front != p->rear) { /* memory is enough. */
#ifdef ADT_DEBUG
		printf("queue_attach(): p->memory=%u is enough, just use it.\n", p->memunits);
#endif
	} else { /* need to realloc() more memory. */
		ptrdiff_t r, f;
		size_t l;
		queue_node_t *t;

		/* prepare pointers for future reposition. */
		r = p->rear - p->memstart, f = p->front - p->memstart, l = p->memunits - f;
#ifdef ADT_DEBUG
		printf("queue_attach(): need to realloc() more memory.\n");
		printf("queue_attach(): r=%u, f=%u, l=%u, p->memunits=%u\n", r, f, l, p->memunits);
#endif
		t = realloc(p->memstart, (p->memunits + QUEUE_INCREMENT) * sizeof(queue_node_t));
		if(!t) {
			/* we don't destory queue if realloc() failed. */
#ifdef ADT_DEBUG
			printf("queue_attach(): realloc() failed but we don't destory queue currently.\n");
#endif
			return false;
		} 

		/* reposition queue. */
#ifdef ADT_DEBUG
		printf("queue_attach(): reposition queue by reposition front node.\n");
#endif
		p->memunits += QUEUE_INCREMENT; 
		p->memstart = t, p->rear = p->memstart + r;
		if(r >= f) {
			p->front = p->memstart + f;
		} else {
			memcpy(t = p->memstart + p->memunits - l, p->front, l);
			p->front = t;
		}
	}
	p->rear->data = data;
	if(++p->rear == p->memstart + p->memunits)
		p->rear = p->memstart;
#ifdef ADT_DEBUG
	{
		size_t i;
		queue_node_t tmpdata, *tq = &tmpdata;
		printf("queue_attach(): p->memstart=%p, p->front=%p, p->rear=%p, p->memunits=%u\n", p->memstart, p->front, p->rear, p->memunits);
#if 0
		for(i=0; i<10; i++) {
			tq = queue_peek(p,i);
			printf("\tpeeking data: %d\n", tq->data);
		}
#endif
	}
#endif
	return true;
#endif
#ifdef ADT_QUEUE_DYNAMIC
	return sll_addtail(p, data);
#endif
#ifdef ADT_QUEUE_STATIC
	if(++p->rear == p->e + QUEUE_MAX_NODES * sizeof(queue_node_t)) {
		p->rear = p->e;
	}
	p->rear->data = data;
	return true;
#endif
}

/* Detach current front from queue.
 * return external storage pointer or NULL on empty queue. */
const void *queue_detach(queue_t *p)
{
#ifdef ADT_QUEUE_BALANCED
	const void *t ;

	/* sanity check. */
	if(p->front == p->rear) {
#ifdef ADT_DEBUG
		printf("queue_detach(): sanity check failed, queue empty!\n");
#endif
		return NULL;
	}

	/* popup front content pointer. */
#ifdef ADT_DEBUG
	printf("queue_detach(): popup front content.\n");
#endif
	t = p->front->data;

	/* reposition queue front. */
#ifdef ADT_DEBUG
	printf("queue_detach(): reposition queue front.\n");
#endif
	if(++p->front == p->memstart + p->memunits) {
		p->front = p->memstart;
	}
	return t;
#endif
#ifdef ADT_QUEUE_DYNAMIC
	const void *t ;

	/* sanity check. */
	if(!p->head) {
		return NULL;
	}

	/* popup front content. */
	t = p->head->data;

	/* reposition queue front. */
	sll_delhead(p);
	return t;
#endif
#ifdef ADT_QUEUE_STATIC
	const void *t ;
	t = p->front->data;
	if(++p->front == p->e + QUEUE_MAX_NODES * sizeof(queue_node_t)) {
		p->front = p->e;
	}
	return t;
#endif
}

/* Peek specific node's element content,
 * return the pointer on valid offset, otherwise NULL. */
const void *queue_peek(queue_t *p, size_t t)
{
#ifdef ADT_QUEUE_BALANCED
	queue_node_t *r = p->front;
	if(t >= queue_getlen(p)) {
#ifdef ADT_DEBUG
		printf("queue_peek(): sanity check failed, out of range!\n");
#endif
		return NULL;
	}
	if(p->rear >= p->front) { /* front before rear. */
#ifdef ADT_DEBUG
		printf("queue_peek(): front before rear\n");
#endif
		return (r+=t);
	} else { /* front after rear, loop happens. */
#ifdef ADT_DEBUG
		printf("queue_peek(): front after rear, loop happens.\n");
#endif
		return r;
	}
#endif
#ifdef ADT_QUEUE_DYNAMIC
	return sll_node_peek(p, t);
#endif
#ifdef ADT_QUEUE_STATIC
	queue_node_t *r = p->e;
	if(t > queue_getlen(p)) {
		return NULL;
	}
	if(p->rear >= p->front) { /* front before rear. */
		r += t;
		return r;
	} else { /* front after rear, loop happens. */
		return r;
	}
#endif
}

/* Get number of node(s) queue contains. */
size_t queue_getlen(queue_t *p)
{
#ifdef ADT_QUEUE_BALANCED
	/* front before rear. */
	if(p->rear >= p->front) {
#ifdef ADT_DEBUG
		printf("queue_getlen(): front before rear, length=%u\n", p->rear - p->front);
#endif
		return p->rear - p->front;
	} else { /* front after rear, loop happens. */
#ifdef ADT_DEBUG
		printf("queue_getlen(): front after rear, loop happens, length=%u\n", p->rear + p->memunits + 1 - p->front);
#endif
		return p->rear + p->memunits + 1 - p->front;
	}
#endif
#ifdef ADT_QUEUE_DYNAMIC
	return p->linklen;
#endif
#ifdef ADT_QUEUE_STATIC
	/* front before rear. */
	if(p->rear >= p->front) {
		return p->rear - p->front;
	} else { /* front after rear, loop happens. */
		return p->rear + QUEUE_MAX_NODES + 1 - p->front;
	}
#endif
}

/* Whether queue is empty(no node) or not. */
bool queue_isempty(queue_t *p)
{
#ifdef ADT_QUEUE_BALANCED
	return p->front != p->rear ? false : true;
#endif
#ifdef ADT_QUEUE_DYNAMIC
	return p->head ? false : true;
#endif
#ifdef ADT_QUEUE_STATIC
	return p->front != p->rear ? false : true;
#endif
}

/* Clear queue content, making it empty/reusable. */
void queue_clear(queue_t *p)
{
#ifdef ADT_QUEUE_BALANCED
	p->rear = p->front = p->memstart;
#endif
#ifdef ADT_QUEUE_DYNAMIC
	sll_destroy(p);
	/* sll_init(p);
	 */
#endif
#ifdef ADT_QUEUE_STATIC
	p->front = p->rear = p->e;
#endif
}

/* Free allocated queue memory. */
void queue_destroy(queue_t *p)
{
#ifdef ADT_QUEUE_BALANCED
	free(p->memstart);
#endif
#ifdef ADT_QUEUE_DYNAMIC
	sll_destroy(p);
#endif
#ifdef ADT_QUEUE_STATIC
#endif
}

#ifdef ADT_QUEUE_DEBUG
int main(void)
{
	queue_t Q, *p = &Q;
	queue_node_t E, *q = &E;
	int i;
	queue_init(p);
	for(i=0; i<10; ++i) {
		q->data = i+100;
		queue_attach(p, q);
	}
	for(i=0; i<10; ++i) {
		queue_detach(p, q);
		printf("data=%d\n", q->data);
	}

	return 0;
}
#endif
