/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of singly linked list using external storage model.
 * see http://en.wikipedia.org/wiki/Linked_list for detail.
 */
#ifndef _SLLIST_H
#define _SLLIST_H

#include <stddef.h>
#include "ulbool.h"
#include "uldef.h"
_BEGIN_DECLS

typedef struct sll_node_s {
	const void *data;		/* const pointer to external storage. */
	struct sll_node_s *next;	/* pointer to next node. */
} sll_node_t;				/* ADT of SLL node. */

typedef struct {
	sll_node_t *head;		/* pointer to head node. */
	sll_node_t *tail;		/* pointer to tail node. */
	size_t linklen;			/* number of nodes SLL contains. */
} sll_t;				/* ADT of SLL. */


/* Init SLL structure to zero. */
void sll_init(sll_t *sll);

/* Duplicate another SLL, pointing to the same external storage. */
bool sll_dup(sll_t *dst, const sll_t *sll);

/* Put SLL content in reversed order. */
void sll_rev(sll_t *r);

/* Delete SLL's current head node. */
void sll_delhead(sll_t *sll);

/* Delete SLL's current tail node. */
void sll_deltail(sll_t *sll);
	
/* Add a new head node. */
bool sll_addhead(sll_t *sll, const void *data);

/* Add a new tail node (append). */
bool sll_addtail(sll_t *sll, const void *data);

/* Conditionally (callback returns true) delete SLL node(s),
 * return number of deleted node(s). */
size_t sll_node_cdel(sll_t *sll, bool (*test)(const void *));

/* Conditionally (callback returns true) update SLL node content,
 * return number of updated node(s). */
size_t sll_node_cupdt(sll_t *sll, bool (*test)(const void *), const void *data);

/* Insert a node right before matching node (callback returns true),
 * return true on success, otherwise false. */
bool sll_node_bins(sll_t *sll, bool (*test)(const void *), const void *data);

/* Insert a node right after matching node (callback returns true),
 * return true on success, otherwise false. */
bool sll_node_ains(sll_t *sll, bool (*test)(const void *), const void *data);

/* Peek specific node's element content,
 * return the pointer on valid offset, otherwise NULL. */
const void *sll_node_peek(sll_t *sll, size_t offset);

/* Number of node(s) SLL contains. */
size_t sll_getlen(sll_t *sll);

/* Sort SLL content by exchanging const pointers to external storage. */
void sll_sort(sll_t *sll, int (*cmp)(const void *, const void *));

/* Merge asc/desc sorted links (all in same order) into a single sorted dst, 
 * various sll_t * parameters and a terminating NULL is expected. 
 * original ones will be destroyed automatically. */
void sll_merge(sll_t *dst, int (*cmp)(const void *, const void *), ...);

/* Traverse SLL via visit() from head to tail. */
void sll_trav(sll_t *dll, void (*visit)(const void *));

/* Free SLL allocated memory. */
void sll_destroy(sll_t *sll);

_END_DECLS
#endif
