/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#if defined(STR_DEBUG) || defined(STR_STRREPL_DEBUG)
#include <stdio.h>
#endif
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "uldef.h"

#define STRREPL_MEMPAGE_SIZE 4096

static size_t used, total;
static char *dst;

static __inline char *cp2malloc(const char *src, size_t len)
{
	char *m;
	size_t d = used + len;
	if(d > total) {	
		do { /* get desired size. */
			total += STRREPL_MEMPAGE_SIZE;
		} while(d > total);
		#ifdef STR_DEBUG
			printf("cp2malloc(): realloc() to new size: %u\n", (unsigned)(total));
		#endif
		dst = realloc(m=dst, total);
		if(!dst) {
			#ifdef STR_DEBUG
				puts("cp2malloc(): realloc() failed !");
			#endif
			free(m); 
			return NULL;
		}
	}
	#ifdef STR_DEBUG
		puts("cp2malloc(): data copying...");
	#endif
	memcpy(dst+used, src, len);
	used += len;
	return dst;
}

/* Replace all matching sub(s) with repl in src, 
 * save result in a newly malloc()ed area,
 * reture a pointer to the area or NULL if malloc() failed.
 * src remain unchanged, don't forget to free()! 
 *
 * fast on large input (O(n)) if realloc() takes constant time, 
 * better choice when strlen(repl) > strlen(sub) many times. */
char *strrepl(const char *p, const char *sub, const char *repl)
{
	char *t;
	size_t ls = strlen(sub), lr;

#ifdef STR_DEBUG
	printf("\nstrrepl(): src=\"%s\"\nstrrepl(): sub=\"%s\"\nstrrepl(): repl=\"%s\"\n", p, sub, repl);
#endif
	/* sanity check. */
	if(ls == 0) {
		#ifdef STR_DEBUG
			printf("strrepl(): strlen(sub)=0, sanity check failed !\n");
		#endif
		goto err;
	}

	/* get data struture ready. */
	used = total = 0, dst = NULL;
	lr = strlen(repl);
	
	/* main loop to check src. */
	while(*p) {
		t = strstr(p,sub);
		if(t) { /* sub found in src. */
			#ifdef STR_DEBUG
				printf("strrepl(): sub=\"%s\"\n", t);
			#endif
			if(p != t) { /* copy segment (between previous sub and current sub) to dst. */
				if(!cp2malloc(p, t-p))
					goto err;
				p = t + ls;
			} else { /* just move pointer forward. */
				p += ls;
			}
			if(!cp2malloc(repl, lr)) /* copy repl to dst. */
				goto err;
		} else { /* only last piece left. */
			#ifdef STR_DEBUG
				printf("strrepl(): last piece=\"%s\"\n", p);
			#endif
			t = cp2malloc(p, strlen(p)+1);
			if(t) { /* we are done. */
				#ifdef STR_DEBUG
					printf("strrepl(): result=\"%s\"\n", t);
					printf("strrepl(): memeory total/used=%u/%u\n", total, used);
				#endif
				return t;
			} else 
				goto err;
		}	
	}
err:
	#ifdef STR_DEBUG
		printf("strrepl(): error exiting...\n");
	#endif
	return NULL;
}

#ifdef STR_STRREPL_DEBUG
int main(int argc, char **argv)
{
	char *buf="hi, this is hcq0411, i like programming, do you like it?";
	char *p;
	
	p=strrepl("", "", "");
	free(p);

	p=strrepl(buf, "hate!!!", "");
	free(p);

	p=strrepl(buf, "", "ABCD");
	free(p);

	p=strrepl(buf, "like", "");
	free(p);

	p=strrepl(buf, "like", "HATE!!!");
	free(p);

	p=strrepl(buf, "like", "LK");
	free(p);

	p=strrepl(buf, "i", "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789876543210~`!@#$%^&*()_+-={}:|<>.,;'[]");
	free(p);

	return 0;
}
#endif
